/*
 * PM - 2017
 * lab0.c
 */

#define F_CPU 16000000

#include <avr/io.h>
#include <util/delay.h>
#include <math.h>

enum Servo {
	BOTTOM_SERVO, // connected on PD4
	TOP_SERVO // connected on PD5
};

enum Direction {
	LEFT,
	RIGHT
};

enum Sensor {
	SENSOR_LEFT1,     // connected on PA0
	SENSOR_LEFT2,     // connected on PA1
	SENSOR_RIGHT1,	  // connected on PA2
	SENSOR_RIGHT2     // connected on PA3
};


const float tracker_speed = 0.5;
float top_servo_angle, bottom_servo_angle;


void pwm_init();
void adc_init();
void tracker_init();

void set_duty_cycle(Servo servo, float f);
void rotate_servo(Servo servo, float angle);
void rotate_tracker(Direction dir);
void set_tracker_angle(float angle);

float read_sensor(Sensor sensor);

int main(void) {
	const float eps = 0.4;
	
	tracker_init();
	
	while(1) {
		float l1 = read_sensor(SENSOR_LEFT1);
		float l2 = read_sensor(SENSOR_LEFT2);
		float r1 = read_sensor(SENSOR_RIGHT1);
		float r2 = read_sensor(SENSOR_RIGHT2);
		
		float left_lum = (l1 + l2) / 2;
		float right_lum = (r1 + r2) / 2;
		
		if(fabs(left_lum - right_lum) < eps)
			continue;
		
		if(left_lum < right_lum)
			rotate_tracker(RIGHT);
		else
			rotate_tracker(LEFT);
		_delay_ms(10);
	}
	
	return 0;
}

void tracker_init() {
	pwm_init();
	adc_init();
	
	top_servo_angle = bottom_servo_angle = 0;
	rotate_servo(TOP_SERVO, top_servo_angle);
	rotate_servo(BOTTOM_SERVO, bottom_servo_angle);
}

void pwm_init() {
	DDRD |= (1 << PD4);
	DDRD |= (1 << PD5);

	// Fast PWM, TOP ICR1
	TCCR1A |= (2 << WGM10);
	TCCR1B |= (3 << WGM12);
	
	// Clear OC1A, OC1B on Compare Match and set at bottom
	TCCR1A |= (1 << COM1A1);
	TCCR1A |= (1 << COM1B1);
		
	// Prescaler of 8
	TCCR1B |= (1 << CS11);
	
	// Frequency of 50 HZ
	ICR1 = 40000;
}


void adc_init() {
	ADMUX = 0;
	ADCSRA = 0;
	
	// AVCC with external capacitor at AREF pin
	ADMUX |= (1 << REFS0);
	
	// clock prescaler at 128
	ADCSRA |= (7 << ADPS0);
	
	// enable ADC
	ADCSRA |= (1 << ADEN);
}

void set_duty_cycle(Servo servo, float f) {
	if(servo == BOTTOM_SERVO)
		OCR1B = f * ICR1;
	else if(servo == TOP_SERVO)
		OCR1A = f * ICR1;
}

void rotate_servo(Servo servo, float angle) {
	const static float duty_cycle_0 = 0.024;
	const static float duty_cycle_180 = 0.117;
	
	float duty_cycle =
		angle / 180 * (duty_cycle_180 - duty_cycle_0) + duty_cycle_0;
	
	set_duty_cycle(servo, duty_cycle);
}

void rotate_tracker(Direction dir) {
	float tracker_angle = top_servo_angle + bottom_servo_angle;
	float new_tracker_angle = tracker_angle + ((dir == LEFT) ? tracker_speed : -tracker_speed);
	float i;
	
	if(new_tracker_angle < 0) {
		for(i = 0; i < 360; i += tracker_speed) {
			set_tracker_angle(i);
			_delay_ms(5);
		}
		new_tracker_angle += 360;
	}
	
	else if(new_tracker_angle > 360) {
		for(i = 360; i > 0; i -= tracker_speed) {
			set_tracker_angle(i);
			_delay_ms(5);
		}
		new_tracker_angle -= 360;
	}

	set_tracker_angle(new_tracker_angle);
}

void set_tracker_angle(float angle) {
	if(angle > 180) {
		bottom_servo_angle = 180;
		top_servo_angle = angle - 180;
	}
	else {
		bottom_servo_angle = angle;
		top_servo_angle = 0;
	}

	rotate_servo(TOP_SERVO, top_servo_angle);
	rotate_servo(BOTTOM_SERVO, bottom_servo_angle);
}

float read_sensor(Sensor sensor) {
	// Select channel
	ADMUX &= ~0x1F;
	switch(sensor) {
		case SENSOR_LEFT1:
			break;
		case SENSOR_LEFT2:
			ADMUX |= (1 << MUX0);
			break;
		case SENSOR_RIGHT1:
			ADMUX |= (2 << MUX0);
			break;
		case SENSOR_RIGHT2:
			ADMUX |= (3 << MUX0);
	}
	
	// read data from sensor
	ADCSRA |= (1 << ADSC);
	while((ADCSRA & (1 << ADSC)));
	
	// convert ADC value to voltage
	int x = ADC;
	float voltage = x * 5 / 1024.0;
	
	return voltage;
}